import shlex
from typing import Any

import click
import questionary
import typer
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import NestedCompleter
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.styles import Style
from questionary import Choice, Separator
from rich.table import Table
from rich.text import Text

from jailrun import cmd
from jailrun.cmd.status import RawJail, get_bastille_jails
from jailrun.qemu import QemuMode, vm_is_running
from jailrun.schemas import State
from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw
from jailrun.ui import COMMANDS, Q_STYLE, con, pick_config, pick_jails_from_config, warn

_SHELL_COMMANDS = [
    ("help", "Show available commands", False),
    ("exit", "Quit", False),
]

_ALIASES: dict[str, str] = {"?": "help", "quit": "exit", "q": "exit"}

PT_STYLE = Style.from_dict({"prompt": "ansicyan bold", "rprompt": "ansibrightblack"})


def _parse(click_app: click.Group, command: str, args: list[str]) -> dict[str, Any]:
    sub_cmd = click_app.commands.get(command)
    if sub_cmd is None:
        return {}
    try:
        with sub_cmd.make_context(command, list(args), resilient_parsing=True) as ctx:
            return dict(ctx.params)
    except Exception:  # noqa: BLE001
        return {}


def _invoke(click_app: click.Group, argv: list[str]) -> None:
    try:
        with click_app.make_context("jrun", argv) as ctx:
            click_app.invoke(ctx)
    except SystemExit:
        pass
    except click.exceptions.Abort:
        _nl()
    except click.exceptions.UsageError as exc:
        con().print(f"\n  [bold red]error:[/bold red] {exc}\n")


def _build_completer(click_app: click.Group) -> NestedCompleter:
    completions: dict[str, Any] = {}
    for name, sub in click_app.commands.items():
        opts = {
            opt: None
            for param in sub.params
            if isinstance(param, click.Option)
            for opt in param.opts
            if opt.startswith("--")
        }
        completions[name] = opts or None

    completions["help"] = None
    completions["?"] = None
    completions["exit"] = None

    return NestedCompleter.from_nested_dict(completions)


def _command_table(include_shell_extras: bool = False) -> Table:
    tbl = Table.grid(padding=(0, 4))
    tbl.add_column(no_wrap=True, min_width=10)
    tbl.add_column(style="dim white")

    for cmd_ in COMMANDS:
        style = "bold red" if cmd_.danger else "bold cyan"
        tbl.add_row(Text(f"  {cmd_.name}", style=style), cmd_.desc)

    if include_shell_extras:
        for name, desc, danger in _SHELL_COMMANDS:
            style = "bold red" if danger else "bold cyan"
            tbl.add_row(Text(f"  {name}", style=style), desc)

    return tbl


def _print_welcome(version: str) -> None:
    c = con()
    c.print()

    logo = Text()
    logo.append("█ ", style="bold cyan")
    logo.append("jrun", style="bold white")
    logo.append(f"  {version}", style="dim white")

    c.print(logo)
    c.print()
    c.print(Text("  Effortless orchestration for FreeBSD jails", style="dim white"))
    c.print()
    c.print(_command_table(include_shell_extras=False))
    c.print()
    c.print(Text("  Tab · /  browse    ↑↓  history    ?  help    Ctrl-C  exit", style="dim cyan"))
    c.print()


def _print_help() -> None:
    c = con()
    c.print()
    c.print(_command_table(include_shell_extras=True))
    c.print()


def _nl() -> None:
    con().print()


def _aborted() -> None:
    warn("Aborted.")


def _fetch_live_jails(state: State, settings: Settings) -> list[RawJail]:
    try:
        ssh_kw = get_ssh_kw(state=state, settings=settings)
        return get_bastille_jails(**ssh_kw)
    except Exception:  # noqa: BLE001
        return []


def _pick_jail_interactively(
    state: State, settings: Settings, *, prompt: str = "SSH into:", allow_host: bool = True
) -> str | None:
    with con().status("[dim]Fetching jail list…[/dim]", spinner="dots"):
        jails = _fetch_live_jails(state=state, settings=settings)

    _nl()

    choices: list[Choice] = []
    if jails:
        max_name = max(len(j["name"]) for j in jails)
        for j in jails:
            state_col = "up" if j["state"].lower() == "up" else j["state"].lower()
            label = f"{j['name'].ljust(max_name)}   {state_col}   {j['ip']}"
            choices.append(Choice(label, value=j["name"]))
        choices.append(Separator())

    if allow_host:
        choices += [Choice("Host VM", value="__host__"), Separator()]

    choices.append(Choice("Cancel", value="__cancel__"))
    chosen = questionary.select(prompt, choices=choices, style=Q_STYLE).ask()

    _nl()

    if chosen in (None, "__cancel__"):
        raise typer.Abort()
    if chosen == "__host__":
        return None

    return str(chosen)


def _is_vm_running(settings: Settings) -> bool:
    alive, _ = vm_is_running(settings.pid_file)
    return alive


def _offer_start_vm(state: State, settings: Settings) -> bool:
    _nl()
    answer = questionary.confirm("VM is not running. Boot it now?", default=True, style=Q_STYLE).ask()
    _nl()

    if not answer:
        return False

    cmd.start(base_config=None, state=state, settings=settings, mode=QemuMode.SERVER)

    return True


def _preflight_ssh(click_app: click.Group, args: list[str], state: State, settings: Settings) -> list[str] | None:
    p = _parse(click_app, "ssh", args)
    if p.get("jail_name") is not None:
        return args

    if not _is_vm_running(settings):
        if not _offer_start_vm(state, settings):
            return None

    jail_name = _pick_jail_interactively(state, settings)

    return args if jail_name is None else [jail_name, *args]


def _preflight_cmd(click_app: click.Group, args: list[str], state: State, settings: Settings) -> list[str] | None:
    p = _parse(click_app, "cmd", args)

    if p.get("jail_name") is not None and p.get("executable") is not None:
        return args

    if not _is_vm_running(settings):
        if not _offer_start_vm(state, settings):
            return None

    jail_name = p.get("jail_name")
    if jail_name is None:
        jail_name = _pick_jail_interactively(state, settings, prompt="Run command in:", allow_host=False)
        if jail_name is None:
            return None

    raw = questionary.text("Executable (with args):", style=Q_STYLE).ask()
    if not raw:
        return None
    _nl()
    try:
        extra = shlex.split(raw)
    except ValueError:
        con().print("\n  [bold red]error:[/bold red] invalid quoting\n")
        return None

    return [jail_name, *extra]


def _preflight_start(args: list[str]) -> list[str] | None:
    if args:
        return args

    _nl()

    want_base = questionary.confirm("Use a custom base.ucl config?", default=False, style=Q_STYLE).ask()
    if want_base is None:
        return None

    _nl()

    if not want_base:
        return args

    raw = questionary.path("Path to base.ucl:", style=Q_STYLE).ask()
    if raw is None:
        return None

    _nl()

    return ["--base", raw] if raw else args


def _preflight_config_required(click_app: click.Group, command: str, args: list[str]) -> list[str] | None:
    p = _parse(click_app, command, args)
    if p.get("config") is not None:
        return args

    config = pick_config()
    if config is None:
        return None

    names = pick_jails_from_config(config)

    _nl()

    flags = [a for a in args if a.startswith("-")]
    result = [str(config), *(names or []), *flags]

    return result


def _preflight(
    click_app: click.Group, command: str, args: list[str], state: State, settings: Settings
) -> list[str] | None:
    if command == "ssh":
        return _preflight_ssh(click_app=click_app, args=args, state=state, settings=settings)
    if command == "cmd":
        return _preflight_cmd(click_app=click_app, args=args, state=state, settings=settings)
    if command == "start":
        return _preflight_start(args)
    if command in ("down", "pause"):
        return _preflight_config_required(click_app=click_app, command=command, args=args)

    return args


def _dispatch(
    state: State,
    settings: Settings,
    click_app: click.Group,
    *,
    token: str,
    inline_args: list[str],
) -> bool:
    c = _ALIASES.get(token.strip().lower(), token.strip().lower())

    if c == "exit":
        return False

    if c == "help":
        _print_help()
        return True

    if c not in click_app.commands:
        con().print(
            f"\n  [bold red]unknown command:[/bold red] [cyan]{c}[/cyan]"
            "  —  type [bold]?[/bold] for help or Tab to browse\n"
        )
        return True

    argv = _preflight(click_app=click_app, command=c, args=inline_args, state=state, settings=settings)

    if argv is None:
        _aborted()
        return True

    _invoke(click_app, [c, *argv])

    return True


def run(state: State, settings: Settings, version: str = "dev", click_app: click.Group | None = None) -> None:
    _print_welcome(version)

    if click_app is None:
        click_app = click.Group()

    completer = _build_completer(click_app)

    kb = KeyBindings()

    @kb.add("/")
    def _slash(event) -> None:  # type: ignore[no-untyped-def]
        buf = event.current_buffer
        if buf.text == "":
            buf.start_completion(select_first=False)
        else:
            buf.insert_text("/")

    session: PromptSession[str] = PromptSession(
        history=InMemoryHistory(),
        completer=completer,
        key_bindings=kb,
        style=PT_STYLE,
        complete_while_typing=False,
        mouse_support=False,
    )

    while True:
        try:
            raw = session.prompt(HTML("<ansicyan><b>jrun</b></ansicyan> <ansibrightblack>›</ansibrightblack> "))
        except KeyboardInterrupt:
            con().print()
            con().print("[dim]  Bye![/dim]\n")
            break
        except EOFError:
            con().print("[dim]\n  Bye![/dim]\n")
            break

        try:
            parts = shlex.split(raw)
        except ValueError:
            con().print("\n  [bold red]error:[/bold red] invalid quoting\n")
            continue

        if not parts:
            continue
        if parts[0] == "?":
            parts[0] = "help"

        try:
            keep_going = _dispatch(
                state=state,
                settings=settings,
                click_app=click_app,
                token=parts[0],
                inline_args=parts[1:],
            )
        except (typer.Exit, typer.Abort):
            keep_going = True
        except Exception as exc:  # noqa: BLE001
            con().print(f"\n  [bold red]error:[/bold red] {exc}\n")
            keep_going = True

        if not keep_going:
            con().print("[dim]  Bye.[/dim]\n")
            break



import contextlib
import json
from pathlib import Path
from typing import NotRequired, TypedDict

import typer
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from jailrun.qemu import vm_is_running
from jailrun.schemas import State
from jailrun.serializers import loads
from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw, ssh_exec, wait_for_ssh
from jailrun.ui import con


class RawJail(TypedDict):
    name: str
    state: str
    ip: str


class JailRow(TypedDict):
    name: str
    state: str
    ip: str
    managed: bool
    ports: str
    mounts: str
    stale: NotRequired[bool]


class StatusInfo(TypedDict):
    pid: int
    ssh_host: str
    ssh_port: int
    uptime: str | None
    disk_free: str | None
    disk_total: str | None
    mem_total: float | None
    mem_usable: float | None
    jail_rows: list[JailRow]


class DiskStats(TypedDict):
    disk_free: str | None
    disk_total: str | None


class MemStats(TypedDict):
    mem_total: float | None
    mem_usable: float | None


def short_path(p: str) -> str:
    parts = Path(p).parts
    if len(parts) > 2:
        return "…/" + str(Path(*parts[-2:]))
    return p


def get_bastille_jails(private_key: Path, ssh_user: str, ssh_host: str, ssh_port: int) -> list[RawJail]:
    raw_jails: list[dict[str, str]] = []
    clean_jails: list[RawJail] = []

    bastille_list = ssh_exec(
        cmd="bastille list -j",
        private_key=private_key,
        ssh_user=ssh_user,
        ssh_host=ssh_host,
        ssh_port=ssh_port,
    )
    if not bastille_list or bastille_list == "[]":
        return []

    with contextlib.suppress(json.JSONDecodeError, TypeError):
        raw_jails = loads(bastille_list)

    for j in sorted(raw_jails, key=lambda x: x.get("Name", "")):
        clean_jails.append(
            RawJail(
                name=j.get("Name", "?"),
                state=j.get("State", "?"),
                ip=j.get("IP Address", "-"),
            )
        )
    return clean_jails


def get_disk_stats(private_key: Path, ssh_user: str, ssh_host: str, ssh_port: int) -> DiskStats:
    disk_free: str | None = None
    disk_total: str | None = None
    disk = ssh_exec(cmd="df -h /", private_key=private_key, ssh_user=ssh_user, ssh_host=ssh_host, ssh_port=ssh_port)
    if disk:
        parts = disk.splitlines()[-1].split()
        if len(parts) >= 4:
            disk_free, disk_total = parts[3], parts[1]

    return DiskStats(disk_free=disk_free, disk_total=disk_total)


def get_mem_stats(private_key: Path, ssh_user: str, ssh_host: str, ssh_port: int) -> MemStats:
    mem_total: float | None = None
    mem_usable: float | None = None
    mem = ssh_exec(
        cmd="sysctl -n hw.physmem hw.usermem",
        private_key=private_key,
        ssh_host=ssh_host,
        ssh_user=ssh_user,
        ssh_port=ssh_port,
    )
    if mem:
        lines = mem.splitlines()
        if len(lines) == 2:
            mem_total = int(lines[0]) / (1024**3)
            mem_usable = int(lines[1]) / (1024**3)
    return MemStats(mem_total=mem_total, mem_usable=mem_usable)


def collect_info(settings: Settings, state: State) -> StatusInfo:
    alive, pid = vm_is_running(settings.pid_file)

    if not alive:
        c = con()
        c.print()
        c.print(
            Text.assemble(
                ("● ", "bold yellow"),
                ("VM", "bold white"),
                ("  not running", "yellow"),
            )
        )
        c.print()
        c.print(Text("  Run jrun start to boot it.", style="dim white"))
        c.print()
        raise typer.Exit(0)

    if pid is None:
        raise RuntimeError("VM reported running but PID is missing")

    ssh_kw = get_ssh_kw(settings, state)

    with con().status("[dim]Connecting to VM…[/dim]", spinner="dots"):
        wait_for_ssh(**ssh_kw, silent=True)
        uptime = ssh_exec(cmd="uptime", **ssh_kw)
        disk_stats = get_disk_stats(**ssh_kw)
        mem_stats = get_mem_stats(**ssh_kw)
        bastille_jails = get_bastille_jails(**ssh_kw)

    managed_names = set(state.jails.keys())
    jail_rows: list[JailRow] = []

    for j in bastille_jails:
        j_name = j["name"]
        managed = j_name in managed_names
        if managed:
            jail_state = state.jails[j_name]
            ports = ", ".join(f"{f.proto}/{f.host}→{f.jail}" for f in jail_state.forwards.values())
            mounts = ", ".join(f"{short_path(m.host)} → {m.jail}" for m in jail_state.mounts.values())
        else:
            ports = ""
            mounts = ""

        jail_rows.append(
            JailRow(
                name=j_name,
                state=j["state"],
                ip=j["ip"],
                managed=managed,
                ports=ports,
                mounts=mounts,
            )
        )

    bastille_names = {j["name"] for j in bastille_jails}
    for name in sorted(managed_names - bastille_names):
        jail = state.jails[name]
        ports = ", ".join(f"{f.proto}/{f.host}→{f.jail}" for f in jail.forwards.values())
        mounts = ", ".join(f"{short_path(m.host)} → {m.jail}" for m in jail.mounts.values())
        jail_rows.append(
            JailRow(
                name=name,
                state="Missing",
                ip=jail.ip or "-",
                managed=True,
                stale=True,
                ports=ports,
                mounts=mounts,
            )
        )

    return StatusInfo(
        pid=int(pid),
        ssh_host=ssh_kw["ssh_host"],
        ssh_port=ssh_kw["ssh_port"],
        uptime=uptime,
        disk_free=disk_stats["disk_free"],
        disk_total=disk_stats["disk_total"],
        mem_total=mem_stats["mem_total"],
        mem_usable=mem_stats["mem_usable"],
        jail_rows=jail_rows,
    )


def _jail_labels(j: JailRow) -> tuple[Text, Text]:
    name, st = j["name"], j["state"]
    if j.get("stale"):
        return (
            Text.assemble((name, "bold"), ("  stale", "yellow")),
            Text("missing", style="yellow"),
        )
    state_style = "green" if st.lower() == "up" else "red"
    if not j["managed"]:
        return (
            Text.assemble((name, "bold"), ("  unmanaged", "dim white")),
            Text(st.lower(), style=state_style),
        )
    return Text(name, style="bold"), Text(st.lower(), style=state_style)


def render_table(data: StatusInfo) -> None:
    c = con()
    c.print()

    c.print(
        Text.assemble(
            ("  ● ", "bold green"),
            ("VM", "bold white"),
            ("  running", "green"),
            (f"  on {data['ssh_host']}:{data['ssh_port']}", "dim white"),
            (f"  (pid {data['pid']})", "dim white"),
        )
    )
    c.print()

    vitals = Table.grid(padding=(0, 3, 0, 0))
    vitals.add_column(style="dim cyan", no_wrap=True, min_width=8)
    vitals.add_column()

    if data["uptime"]:
        vitals.add_row("uptime", data["uptime"].strip())
    if data["disk_free"] and data["disk_total"]:
        vitals.add_row(
            "disk",
            Text.assemble((data["disk_free"], "bold green"), (f" free of {data['disk_total']}", "dim white")),
        )
    if data["mem_total"] is not None and data["mem_usable"] is not None:
        vitals.add_row(
            "memory",
            Text.assemble(
                (f"{data['mem_usable']:.1f} GB", "bold green"),
                (f" usable / {data['mem_total']:.1f} GB total", "dim white"),
            ),
        )
    c.print(Padding(vitals, pad=(0, 0, 0, 2)))
    c.print()

    if not data["jail_rows"]:
        c.print("  [dim]no jails[/dim]")
        c.print()
        return

    tbl = Table(
        show_header=True,
        header_style="dim",
        box=None,
        show_edge=False,
        padding=(0, 3, 0, 0),
        expand=False,
        pad_edge=False,
    )
    tbl.add_column("name", no_wrap=True)
    tbl.add_column("state", no_wrap=True)
    tbl.add_column("ip", style="dim white", no_wrap=True)
    tbl.add_column("ports", style="dim white")
    tbl.add_column("mounts", style="dim white")

    for j in data["jail_rows"]:
        name_t, state_t = _jail_labels(j)
        tbl.add_row(
            name_t,
            state_t,
            j["ip"],
            j["ports"] or Text("—", style="dim white"),
            j["mounts"] or Text("—", style="dim white"),
        )

    c.print(Padding(tbl, pad=(0, 0, 0, 2)))
    c.print()


def render_tree(data: StatusInfo) -> None:
    c = con()
    c.print()

    root = Tree(
        Text.assemble(
            ("● ", "bold green"),
            ("VM", "bold white"),
            ("  running", "green"),
            (f"  on {data['ssh_host']}:{data['ssh_port']}", "dim white"),
            (f"  (pid {data['pid']})", "dim white"),
        )
    )

    if data["uptime"]:
        root.add(Text.assemble(("uptime  ", "dim cyan"), data["uptime"].strip()))
    if data["disk_free"] and data["disk_total"]:
        root.add(
            Text.assemble(
                ("disk    ", "dim cyan"),
                (data["disk_free"], "bold green"),
                (f" free of {data['disk_total']}", "dim white"),
            )
        )
    if data["mem_total"] is not None and data["mem_usable"] is not None:
        root.add(
            Text.assemble(
                ("memory  ", "dim cyan"),
                (f"{data['mem_usable']:.1f} GB", "bold green"),
                (f" usable / {data['mem_total']:.1f} GB total", "dim white"),
            )
        )

    jails_node = root.add(Text("jails", style="bold"))

    if not data["jail_rows"]:
        jails_node.add(Text("no jails", style="dim white"))
    else:
        for j in data["jail_rows"]:
            name_t, state_t = _jail_labels(j)
            label = name_t + Text("  ") + state_t + Text(f"  {j['ip']}", style="dim white")
            node = jails_node.add(label)
            node.add(Text.assemble(("ports   ", "dim cyan"), j["ports"] or Text("—", style="dim white")))
            node.add(Text.assemble(("mounts  ", "dim cyan"), j["mounts"] or Text("—", style="dim white")))

    c.print(Padding(root, pad=(0, 0, 0, 2)))
    c.print()


def status(state: State, settings: Settings, *, tree: bool = False) -> None:
    data = collect_info(state=state, settings=settings)
    if tree:
        render_tree(data)
    else:
        render_table(data)



from pathlib import Path

import typer

from jailrun.ansible import run_playbook
from jailrun.config import parse_config
from jailrun.misc import lock
from jailrun.qemu import vm_is_running
from jailrun.schemas import JailPlan, Plan, State
from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw, wait_for_ssh
from jailrun.ui import err, ok, warn


def pause(config: Path, state: State, settings: Settings, *, names: list[str] | None = None) -> None:
    with lock(settings.state_file):
        _pause(config=config, state=state, settings=settings, names=names)


def _pause(config: Path, state: State, settings: Settings, *, names: list[str] | None = None) -> None:
    cfg = parse_config(config)
    targets = set(names) if names else set(cfg.jail.keys())

    unknown = targets - set(cfg.jail.keys())
    if unknown:
        warn(f"Not in config: {', '.join(sorted(unknown))}")

    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise typer.Exit(1)

    to_stop: list[str] = []
    for name in sorted(targets):
        if name in state.jails:
            to_stop.append(name)
        else:
            warn(f"Jail '{name}' not in state — skipping.")

    if not to_stop:
        warn("No matching jails found in state.")
        return

    ssh_kw = get_ssh_kw(settings, state)
    wait_for_ssh(**ssh_kw, silent=True)

    stop_plan = Plan(
        jails=[
            JailPlan(
                name=name,
                release=state.jails[name].release,
                ip=state.jails[name].ip,
            )
            for name in to_stop
        ],
    )

    run_playbook("jail-stop.yml", plan=stop_plan, settings=settings, state=state)

    ok(f"Paused: {', '.join(to_stop)}.")



from pathlib import Path
from typing import NamedTuple

import questionary
import typer
from questionary import Choice, Separator
from rich.console import Console

from jailrun import ucl

CYAN = "#00aaaa"


class Command(NamedTuple):
    name: str
    desc: str
    danger: bool = False


COMMANDS: list[Command] = [
    Command("start", "Boot the VM"),
    Command("stop", "Gracefully shut down the VM", danger=True),
    Command("ssh", "Open a shell in the VM or a jail"),
    Command("cmd", "Execute a command inside a jail"),
    Command("up", "Create or update jails from config"),
    Command("down", "Stop and destroy jails", danger=True),
    Command("pause", "Stop jails without destroying them"),
    Command("purge", "Destroy the VM and ALL jails", danger=True),
    Command("status", "Show VM and jail status"),
]


def con() -> Console:
    return Console()


def ok(msg: str) -> None:
    con().print(f"\n[bold green]✓[/bold green] {msg}\n")


def warn(msg: str) -> None:
    con().print(f"[yellow]![/yellow] [dim]{msg}[/dim]\n")


def err(msg: str) -> None:
    con().print(f"[bold red]✗[/bold red] {msg}\n")


def info(msg: str) -> None:
    con().print(f"[dim]{msg}[/dim]")


Q_STYLE = questionary.Style(
    [
        ("qmark", f"fg:{CYAN} bold"),
        ("question", "bold"),
        ("answer", f"fg:{CYAN} bold"),
        ("pointer", f"fg:{CYAN} bold"),
        ("highlighted", f"fg:{CYAN}"),
        ("selected", f"fg:{CYAN}"),
        ("separator", "fg:#444444"),
        ("instruction", "fg:#555555 italic"),
        ("text", ""),
        ("disabled", "fg:#555555 italic"),
    ]
)


def pick_config() -> Path:
    candidates = sorted(Path(".").glob("**/*.ucl"))
    con().print()

    if candidates:
        choices: list[Choice] = [Choice(str(p), value=p) for p in candidates]
        choices += [
            Separator(),
            Choice("Enter path manually…", value="__manual__"),
            Choice("Cancel", value="__cancel__"),
        ]
        chosen = questionary.select(
            "Select jail config:",
            choices=choices,
            style=Q_STYLE,
        ).ask()

        if chosen is None or chosen == "__cancel__":
            raise typer.Abort()

        if chosen == "__manual__":
            raw = questionary.path("Path to config (.ucl):", style=Q_STYLE).ask()
            if not raw:
                raise typer.Abort()
            return Path(raw)
        return Path(chosen)

    raw = questionary.path("Path to jail config (.ucl):", style=Q_STYLE).ask()
    if not raw:
        raise typer.Abort()

    return Path(raw)


def _parse_jail_names_from_ucl(config: Path) -> list[str]:
    ucl_config = ucl.load_file(str(config))
    if "jail" not in ucl_config:
        return []

    return list(ucl_config["jail"].keys())


def pick_jails_from_config(config: Path) -> list[str] | None:
    names = _parse_jail_names_from_ucl(config)

    scope = questionary.select(
        "Target which jails?",
        choices=[
            c
            for c in [
                Choice("All jails", value="__all__"),
                Choice("Choose specific…", value="__pick__") if names else None,
                Separator(),
                Choice("Cancel", value="__cancel__"),
            ]
            if c is not None
        ],
        style=Q_STYLE,
    ).ask()

    if scope is None or scope == "__cancel__":
        raise typer.Abort()
    if scope == "__all__":
        return None

    con().print()

    choices = [Choice(n, value=n, checked=False) for n in names]

    selected = questionary.checkbox(
        "Select jails:",
        choices=choices,
        style=Q_STYLE,
    ).ask()

    if selected is None or not selected:
        raise typer.Abort()

    return list(selected)



import os
import platform
import shutil
import subprocess
import tempfile
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path

from jailrun.config import derive_qemu_fwds, derive_qemu_shares
from jailrun.http import download
from jailrun.schemas import QemuFwd, QemuShare, State
from jailrun.serializers import loads
from jailrun.settings import Settings
from jailrun.ssh import ensure_vm_key, find_free_port, is_port_free
from jailrun.templates import build_jinja_env
from jailrun.ui import info, ok, warn


@dataclass(frozen=True)
class QemuFeatures:
    arch: str
    machine: str
    accel: str
    bios: str
    virtio_suffix: str
    display: str


class QemuMode(StrEnum):
    SERVER = "server"
    TTY = "tty"
    GRAPHIC = "graphic"


def detect_qemu_features() -> QemuFeatures:
    system = platform.system().lower()
    machine = platform.machine()

    if system == "darwin":
        if machine == "arm64":
            return QemuFeatures(
                arch="aarch64",
                machine="virt",
                accel="hvf",
                bios="edk2-aarch64-code.fd",
                virtio_suffix="device",
                display="cocoa",
            )
        return QemuFeatures(
            arch="x86_64",
            machine="q35",
            accel="hvf",
            bios="/usr/local/share/qemu/edk2-x86_64-code.fd",
            virtio_suffix="pci",
            display="cocoa",
        )

    if machine == "aarch64":
        return QemuFeatures(
            arch="aarch64",
            machine="virt",
            accel="kvm",
            bios="/usr/share/AAVMF/AAVMF_CODE.fd",
            virtio_suffix="device",
            display="gtk",
        )

    return QemuFeatures(
        arch="x86_64",
        machine="q35",
        accel="kvm",
        bios="/usr/share/OVMF/OVMF_CODE.fd",
        virtio_suffix="pci",
        display="gtk",
    )


def parse_size(value: str) -> int:
    units = {"K": 1024, "M": 1024**2, "G": 1024**3, "T": 1024**4}
    v = value.strip().upper()
    if v[-1].isdigit():
        return int(v)

    return int(float(v[:-1]) * units[v[-1]])


def build_netdev_arg(hostfwd: list[QemuFwd], *, ssh_host: str, ssh_port: int) -> str:
    parts = ["user", "id=net0", f"hostfwd=tcp:{ssh_host}:{ssh_port}-:22"]
    for rule in hostfwd:
        parts.append(f"hostfwd={rule.proto}:{ssh_host}:{rule.host}-:{rule.guest}")

    return ",".join(parts)


def build_share_args(shares: list[QemuShare], *, features: QemuFeatures) -> list[str]:
    args: list[str] = []
    for share in shares:
        args += [
            "-fsdev",
            f"local,id={share.id},path={share.host},security_model=none,readonly=off",
            "-device",
            f"virtio-9p-{features.virtio_suffix},fsdev={share.id},mount_tag={share.mount_tag}",
        ]
    return args


def disk_size_bytes(path: Path) -> int:
    out = subprocess.run(
        ["qemu-img", "info", "--output=json", str(path)],
        capture_output=True,
        text=True,
        check=True,
    )
    return int(loads(out.stdout)["virtual-size"])


def build_qemu_cmd(state: State, *, settings: Settings, mode: QemuMode) -> list[str]:
    if state.ssh_port is None:
        raise RuntimeError("state.ssh_port is not set")

    image_xz = Path(str(settings.bsd_image_url)).name
    disk_path = settings.disk_dir / Path(image_xz).with_suffix("")
    cloud_iso = settings.cloud_dir / "cloud-init.iso"

    features = detect_qemu_features()

    hostfwd = derive_qemu_fwds(state=state)
    shares = derive_qemu_shares(state)

    netdev = build_netdev_arg(hostfwd=hostfwd, ssh_host=settings.vm_host, ssh_port=state.ssh_port)
    share_args = build_share_args(shares, features=features)

    cmd = [
        f"qemu-system-{features.arch}",
        "-m",
        settings.qemu_memory,
        "-M",
        f"{features.machine},accel={features.accel}",
        "-cpu",
        "host",
        "-bios",
        features.bios,
        "-device",
        f"virtio-net-{features.virtio_suffix},netdev=net0",
        "-netdev",
        netdev,
        "-drive",
        f"if=virtio,file={disk_path},format=raw,cache=writethrough",
        "-drive",
        f"file={cloud_iso},media=cdrom",
        *share_args,
    ]

    if mode == QemuMode.SERVER:
        cmd.extend(["-display", "none"])
    if mode == QemuMode.TTY:
        cmd.extend(["-nographic", "-serial", "mon:stdio"])
    if mode == QemuMode.GRAPHIC:
        cmd.extend(
            [
                "-display",
                features.display,
                "-device",
                "ramfb",
                "-device",
                "qemu-xhci",
                "-device",
                "usb-kbd",
                "-device",
                "usb-tablet",
            ]
        )

    return cmd


def resolve_ssh_port(state: State, *, settings: Settings) -> int:
    if state.ssh_port is None:
        state.ssh_port = find_free_port(settings.ssh_port, bind_addr=settings.vm_host)
    elif not is_port_free(state.ssh_port, bind_addr=settings.vm_host):
        new_port = find_free_port(state.ssh_port, bind_addr=settings.vm_host)
        warn(f"Previously used SSH port {state.ssh_port} is now busy — switching to {new_port}…")
        state.ssh_port = new_port

    return state.ssh_port


def launch_vm(state: State, *, mode: QemuMode, settings: Settings) -> int | None:
    if state.ssh_port is None:
        raise RuntimeError("state.ssh_port is not set")

    cmd = build_qemu_cmd(state=state, mode=mode, settings=settings)

    if mode in {QemuMode.TTY, QemuMode.GRAPHIC}:
        info(f"Starting VM in {mode} mode…")
        settings.pid_file.unlink(missing_ok=True)
        subprocess.run(cmd, check=False)
        return None

    log_file = settings.log_dir / "qemu.log"
    info("Starting VM in background…")

    with open(log_file, "ab") as log:
        proc = subprocess.Popen(cmd, stdout=log, stderr=log, start_new_session=True)

    settings.pid_file.write_text(str(proc.pid))

    try:
        proc.wait(timeout=3)
        settings.pid_file.unlink(missing_ok=True)
        raise RuntimeError(f"QEMU process {proc.pid} exited immediately — check {log_file}")
    except subprocess.TimeoutExpired:
        pass

    ok(f"VM started on {settings.vm_host}:{state.ssh_port} (pid {proc.pid}).")

    return proc.pid


def prepare_disk(settings: Settings) -> None:
    download(
        bsd_image_url=str(settings.bsd_image_url),
        bsd_image_checksum_url=str(settings.bsd_image_checksum_url),
        target_dir=settings.disk_dir,
    )

    image_xz = Path(str(settings.bsd_image_url)).name
    disk_path = settings.disk_dir / Path(image_xz).with_suffix("")
    xz_path = settings.disk_dir / image_xz

    if not disk_path.exists():
        info("Decompressing image…")
        subprocess.run(["xz", "-dk", str(xz_path)], check=True)

    target = parse_size(settings.qemu_disk_size)
    current = disk_size_bytes(disk_path)

    prepare_cloud_init(settings)
    build_cloud_iso(settings)

    if current < target:
        info(f"Resizing disk to {settings.qemu_disk_size}…")
        subprocess.run(
            ["qemu-img", "resize", "-f", "raw", str(disk_path), settings.qemu_disk_size],
            check=True,
        )


def vm_is_running(pid_file: Path) -> tuple[bool, int | None]:
    if not pid_file.exists():
        return False, None

    pid = int(pid_file.read_text().strip())

    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        pid_file.unlink(missing_ok=True)
        return False, None
    except PermissionError:
        return True, pid

    try:
        out = subprocess.run(
            ["ps", "-p", str(pid), "-o", "comm="],
            capture_output=True,
            text=True,
        )
        if "qemu" not in out.stdout.lower():
            pid_file.unlink(missing_ok=True)
            return False, None
    except Exception:
        pass

    return True, pid


def prepare_cloud_init(settings: Settings) -> None:
    ssh_key = ensure_vm_key(
        private_key=settings.ssh_dir / settings.ssh_key,
        public_key=settings.ssh_dir / f"{settings.ssh_key}.pub",
    )
    env = build_jinja_env()

    user_data = env.get_template("cloud_user_data.j2").render(ssh_key=ssh_key)
    meta_data = env.get_template("cloud_meta_data.j2").render()

    (settings.cloud_dir / "user-data").write_text(f"#cloud-config\n{user_data}")
    (settings.cloud_dir / "meta-data").write_text(meta_data)


def build_cloud_iso(settings: Settings) -> None:
    iso = settings.cloud_dir / "cloud-init.iso"
    with tempfile.TemporaryDirectory() as tmp:
        tmp_path = Path(tmp)
        shutil.copy(settings.cloud_dir / "user-data", tmp_path)
        shutil.copy(settings.cloud_dir / "meta-data", tmp_path)
        subprocess.run(
            ["mkisofs", "-output", str(iso), "-volid", "cidata", "-joliet", "-rock", str(tmp_path)],
            check=True,
        )



import os
import subprocess
import tempfile
from pathlib import Path
from typing import Any

import typer

from jailrun import PACKAGE_DIR
from jailrun.schemas import Plan, State
from jailrun.serializers import dumps
from jailrun.settings import Settings
from jailrun.ssh import SSH_OPTS, get_ssh_kw, proxy_cmd
from jailrun.ui import err, info, ok


def resolve_playbook_path(playbook: str | Path) -> Path:
    p = Path(playbook)
    if p.is_absolute():
        return p.resolve()

    builtins_dir = (PACKAGE_DIR / "playbooks").resolve()

    candidate_user = (Path.cwd() / p).resolve()
    if candidate_user.exists():
        return candidate_user

    candidate_builtin = (builtins_dir / p).resolve()
    if candidate_builtin.exists():
        return candidate_builtin

    return candidate_user


def run_playbook(
    name: str,
    *,
    plan: Plan | None = None,
    settings: Settings,
    state: State,
    jail_name: str | None = None,
    jail_ip: str | None = None,
    extra_vars: dict[str, Any] | None = None,
) -> None:
    playbook = resolve_playbook_path(name)
    if not playbook.exists():
        err(f"Missing playbook: {playbook}")
        raise typer.Exit(1)

    env = os.environ.copy()
    env["ANSIBLE_HOST_KEY_CHECKING"] = "False"

    ssh_kw = get_ssh_kw(settings, state)

    vars = {
        "ansible_host": ssh_kw["ssh_host"],
        "ansible_port": ssh_kw["ssh_port"],
        "ansible_user": ssh_kw["ssh_user"],
        "ansible_python_interpreter": settings.vm_python_interpreter,
        "ansible_ssh_common_args": " ".join(SSH_OPTS),
        "bsd_version": settings.bsd_version,
        "bsd_release_tag": settings.bsd_release_tag,
    }

    if jail_name and jail_ip:
        proxy = proxy_cmd(
            private_key=ssh_kw["private_key"],
            ssh_host=ssh_kw["ssh_host"],
            ssh_user=ssh_kw["ssh_user"],
            ssh_port=ssh_kw["ssh_port"],
        )
        vars.update(
            {
                "ansible_host": jail_ip,
                "ansible_port": 22,
                "ansible_user": "root",
                "ansible_ssh_common_args": f'{" ".join(SSH_OPTS)} -o ProxyCommand="{proxy}"',
            }
        )

    if extra_vars:
        vars.update(extra_vars)

    cmd = [
        "ansible-playbook",
        "-i",
        "localhost,",
        "-c",
        "ssh",
        str(playbook),
        "--private-key",
        str(ssh_kw["private_key"]),
    ]
    cmd += ["-e", dumps(vars)]

    if plan:
        with tempfile.NamedTemporaryFile(suffix=".json", prefix="jrun-plan-", mode="w") as f:
            f.write(plan.model_dump_json(indent=2))
            f.flush()
            cmd += ["-e", f"@{f.name}"]
            info(f"Running playbook {name}…")
            subprocess.run(cmd, check=True, env=env)
    else:
        info(f"Running playbook {name}…")
        subprocess.run(cmd, check=True, env=env)

    ok(f"Playbook {name} complete.")



from pathlib import Path
from types import TracebackType
from typing import Self

from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw, jail_ssh_exec
from jailrun.testing.commons import Jail


class MySQLJail(Jail):
    def __init__(
        self,
        jail: str = "hypha-mysql-test",
        *,
        jail_config: Path,
        base_config: Path | None = None,
        settings: Settings | None = None,
        user: str = "test",
        password: str = "test",
        dbname: str = "testdb",
        port: int = 4306,
    ) -> None:
        self.user = user
        self.password = password
        self.dbname = dbname
        self.port = port
        super().__init__(jail=jail, jail_config=jail_config, base_config=base_config, settings=settings)

    def is_ready(self) -> bool:
        result = jail_ssh_exec(
            "mysqladmin -u root ping",
            jail_ip=self._jail_ip,
            **get_ssh_kw(self._settings, self._state),
        )
        return result is not None

    def __enter__(self) -> Self:
        jail_ssh_exec(
            f"mysql -u root -e 'DROP DATABASE IF EXISTS `{self.dbname}`'",
            jail_ip=self._jail_ip,
            **get_ssh_kw(self._settings, self._state),
        )
        jail_ssh_exec(
            f"mysql -u root -e 'CREATE DATABASE `{self.dbname}`'",
            jail_ip=self._jail_ip,
            **get_ssh_kw(self._settings, self._state),
        )
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        jail_ssh_exec(
            f"mysql -u root -e 'DROP DATABASE IF EXISTS `{self.dbname}`'",
            jail_ip=self._jail_ip,
            **get_ssh_kw(self._settings, self._state),
        )



from jailrun.cmd.down import down as down
from jailrun.cmd.pause import pause as pause
from jailrun.cmd.purge import purge as purge
from jailrun.cmd.ssh import run_cmd as run_cmd
from jailrun.cmd.ssh import ssh as ssh
from jailrun.cmd.start import start as start
from jailrun.cmd.status import status as status
from jailrun.cmd.stop import stop as stop
from jailrun.cmd.up import up as up



from pathlib import Path

import typer

from jailrun.ansible import run_playbook
from jailrun.config import (
    derive_plan,
    parse_config,
    save_state,
)
from jailrun.misc import lock
from jailrun.qemu import vm_is_running
from jailrun.schemas import State
from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw, wait_for_ssh
from jailrun.ui import err, ok, warn


def down(config: Path, state: State, settings: Settings, *, names: list[str] | None = None) -> None:
    with lock(settings.state_file):
        _down(config=config, state=state, settings=settings, names=names)


def _down(config: Path, state: State, *, settings: Settings, names: list[str] | None = None) -> None:
    cfg = parse_config(config)
    targets = set(names) if names else set(cfg.jail.keys())

    unknown = targets - set(cfg.jail.keys())
    if unknown:
        warn(f"Not in config: {', '.join(sorted(unknown))}")

    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise typer.Exit(1)

    new_state = state.model_copy(deep=True)

    removed: list[str] = []
    for name in sorted(targets):
        if name in new_state.jails:
            del new_state.jails[name]
            removed.append(name)
        else:
            warn(f"Jail '{name}' not in state — skipping.")

    if not removed:
        warn("No matching jails found in state.")
        return

    plan = derive_plan(state, new_state)

    ssh_kw = get_ssh_kw(settings, new_state)
    wait_for_ssh(**ssh_kw, silent=True)

    run_playbook("jail-teardown.yml", plan=plan, settings=settings, state=new_state)
    run_playbook("vm-mounts.yml", plan=plan, settings=settings, state=new_state)
    run_playbook("jail-forwards.yml", plan=plan, settings=settings, state=new_state)

    save_state(state=new_state, state_file=settings.state_file)

    ok(f"Removed: {', '.join(removed)}.")



from pathlib import Path

import typer

from jailrun.ansible import run_playbook
from jailrun.cmd.stop import stop_vm
from jailrun.config import (
    derive_plan,
    load_base_into_state,
    save_state,
    snapshot_qemu_wiring,
)
from jailrun.misc import lock
from jailrun.qemu import QemuMode, launch_vm, prepare_disk, resolve_ssh_port, vm_is_running
from jailrun.remote import fetch_remote_playbook
from jailrun.schemas import LocalSetupStep, RemoteSetupStep, State
from jailrun.settings import Settings
from jailrun.ssh import get_ssh_kw, wait_for_ssh
from jailrun.ui import info, warn


def start(
    base_config: Path | None,
    state: State,
    settings: Settings,
    *,
    provision: bool = False,
    mode: QemuMode = QemuMode.SERVER,
) -> None:
    with lock(settings.state_file):
        _start_vm(
            base_config=base_config,
            state=state,
            settings=settings,
            provision=provision,
            mode=mode,
        )


def _start_vm(
    base_config: Path | None,
    state: State,
    settings: Settings,
    *,
    provision: bool = False,
    mode: QemuMode = QemuMode.SERVER,
) -> None:
    alive, pid = vm_is_running(settings.pid_file)
    if alive:
        warn(f"VM is already running (pid {pid}).")
        raise typer.Exit(0)

    prepare_disk(settings)

    needs_base = not settings.state_file.exists()

    new_state = state.model_copy(deep=True)
    new_state = load_base_into_state(base_config, new_state)

    resolve_ssh_port(state=new_state, settings=settings)
    launch_vm(state=new_state, mode=QemuMode.SERVER, settings=settings)
    snapshot_qemu_wiring(state=new_state)
    save_state(state=new_state, state_file=settings.state_file)

    ssh_kw = get_ssh_kw(settings, new_state)
    wait_for_ssh(**ssh_kw)

    if needs_base or provision:
        run_playbook("base.yml", settings=settings, state=new_state)

    for step in new_state.base.setup.values():
        if step.type == "ansible":
            if isinstance(step, RemoteSetupStep):
                playbook_path = fetch_remote_playbook(
                    step.url,
                    cache_dir=settings.playbook_cache_dir,
                )
                run_playbook(
                    str(playbook_path),
                    extra_vars=step.vars or None,
                    settings=settings,
                    state=new_state,
                )
            if isinstance(step, LocalSetupStep):
                run_playbook(
                    step.file,
                    extra_vars=step.vars or None,
                    settings=settings,
                    state=new_state,
                )

    if new_state.base.mounts or new_state.base.forwards:
        plan = derive_plan(state, new_state)
        run_playbook("vm-mounts.yml", plan=plan, settings=settings, state=new_state)

    save_state(state=new_state, state_file=settings.state_file)

    if mode in {QemuMode.TTY, QemuMode.GRAPHIC}:
        info(f"Restarting VM in {mode} mode…")
        stop_vm(settings)
        launch_vm(state=new_state, mode=mode, settings=settings)
